/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include "lay_dns.h"
#include "lay_ip.h"
#include "lay_ether.h"
#include "lay_data.h"
#include "typ_32.h"
#include "typ_16.h"
#include "typ_8.h"
#include "typ_dnsqname.h"
#include "typ_dnsrdata.h"

char				*dnsrttl_itmpl = "\n\
<table _name=dnsrttl[%i%] bgcolor=\"%%dnsrttlColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<a href=\"set(dnsrttl[%i%].ttl)\">%%dnsrttl[%i%].ttl%%</a>\n\
</td>\n\
</tr>\n\
</table>\n\
";

t_field				dnsrttl_fields[] = 
{
  {"ttl",	0,	typ_nu32_msg,			NULL},
  NULL_FIELD
};

t_status			lay_dnsrttl_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_dnsrttl_msg,"dnsrttl");
      LAY_GET_FIELD_GENERIC(dnsrttl_fields);
      LAY_SET_FIELD_GENERIC(dnsrttl_fields);
      LAY_GET_FIELDS_GENERIC(dnsrttl_fields);
      LAY_GET_FIELD_TYP_GENERIC(dnsrttl_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_dnsrttl_msg,dnsrttl_itmpl);
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	
	(*off) = sizeof (t_u32);
	return (0);
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);
	
	(*mp) = &lay_dnsrdata_msg;
	return (0);
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}

char				*dnsrdata_itmpl = "\n\
<table _name=dnsrdata[%i%] bgcolor=\"%%dnsrdataColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<code>\n\
<a href=\"set(dnsrdata[%i%].data)\">%%dnsrdata[%i%].data%%</a>\n\
</code>\n\
</td>\n\
</tr>\n\
</table>\n\
";

t_field				dnsrdata_fields[] = 
{
  {"data",	0,	typ_dnsrdata_msg,		NULL},
  NULL_FIELD
};

t_status			lay_dnsrdata_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_dnsrdata_msg,"dnsrdata");
      LAY_GET_FIELD_GENERIC(dnsrdata_fields);
      LAY_SET_FIELD_GENERIC(dnsrdata_fields);
      LAY_GET_FIELDS_GENERIC(dnsrdata_fields);
      LAY_GET_FIELD_TYP_GENERIC(dnsrdata_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_dnsrdata_msg,dnsrdata_itmpl);
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	t_extract_data		ed;
	
	ed.b.buf = b->buf;
	ed.b.len = b->len;
	ed.data = NULL;
	return (typ_msg(typ_dnsrdata_msg,
			TYP_OFF,
			&ed,
			off));
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);

	(*mp) = lay_dnsqname_msg;
	return (0);
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}

char				*dnsqtype_itmpl = "\n\
<table _name=dnsqtype[%i%] bgcolor=\"%%dnsqtypeColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(dnsqtype[%i%].Type)\">%%dnsqtype[%i%].Type%%</a>(<a href=\"set(dnsqtype[%i%].type)\">%%dnsqtype[%i%].type%%</a>)\n\
</td>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(dnsqtype[%i%].class)\">%%dnsqtype[%i%].class%%</a>\n\
</td>\n\
</tr>\n\
</table>\n\
";

t_assoc				dnsqtype_assocs[] = 
{
  {"a",				(VOID_PTR)1},
  {"ns",			(VOID_PTR)2},
  {"cname",			(VOID_PTR)5},
  {"ptr",			(VOID_PTR)12},
  {"hinfo",			(VOID_PTR)13},
  {"mx",			(VOID_PTR)15},
  {"axfr",			(VOID_PTR)252},
  {"any",			(VOID_PTR)255},
  {NULL,			NULL},
};

t_field				dnsqtype_fields[] = 
{
  {"type",	0,	typ_nu16_msg,			NULL},
  {"Type",	0,	typ_nu16assoc_msg,		dnsqtype_assocs},
  {"class",	2,	typ_nu16_msg,			NULL},
  NULL_FIELD
};

t_status			lay_dnsqtype_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_dnsqtype_msg,"dnsqtype");
      LAY_GET_FIELD_GENERIC(dnsqtype_fields);
      LAY_SET_FIELD_GENERIC(dnsqtype_fields);
      LAY_GET_FIELDS_GENERIC(dnsqtype_fields);
      LAY_GET_FIELD_TYP_GENERIC(dnsqtype_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_dnsqtype_msg,dnsqtype_itmpl);
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	
	(*off) = sizeof (t_u32);
	return (0);
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);
	
	(*mp) = &lay_dnsqname_msg;
	return (0);
      }
    case LAY_SUB_ID:
      {
	LAY_SUB_ID_ARGS(sid,mp);
	t_hash_elt	*he;

	if (he = id_get(sid->id,
			&dns_nquestions_chan))
	  {
	    int		nquestions;
	    t_boolean	again;

	    nquestions = (int)(he->value);
	    if (nquestions == 0)
	      again = FALSE;
	    else
	      {
		again = TRUE;
		nquestions--;
		if ((status = id_override(sid->id,
					  &dns_nquestions_chan,
					  (VOID_PTR)nquestions)) < 0)
		  return (status);
	      }
	    if (again)
	      (*mp) = &lay_dnsqname_msg;
	    else
	      (*mp) = &lay_dnsrttl_msg;
	  }
	else
	  (*mp) = &lay_dnsqname_msg;
	return (0);
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}

char				*dnsqname_itmpl = "\n\
<table _name=dnsqname[%i%] bgcolor=\"%%dnsqnameColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<small>\n\
<a href=\"set(dnsqname[%i%].name)\">%%dnsqname[%i%].name%%</a>\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

t_field				dnsqname_fields[] = 
{
  {"name",	0,	typ_dnsqname_msg,		NULL},
  NULL_FIELD
};

t_status			lay_dnsqname_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_dnsqname_msg,"dnsqname");
      LAY_SET_FIELD_GENERIC(dnsqname_fields);
      LAY_GET_FIELDS_GENERIC(dnsqname_fields);
      LAY_GET_FIELD_TYP_GENERIC(dnsqname_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_dnsqname_msg,dnsqname_itmpl);
    case LAY_GET_FIELD:
      {
	LAY_GET_FIELD_ARGS(gfd,bs);

	if (!strcmp(gfd->field,"name"))
	  {
	    int			len;
	    t_extract_data	ed;

	    if (gfd->b.len < 1)
	      return (-ERR_LAYER_TRUNC);
	    len = (t_u32)(t_u8)(gfd->b.buf[0]);
	    if ((len & INDIR_MASK) == INDIR_MASK)
	      return (-ERR_LAYER_CANTHANDLE);
	    ed.b.buf = gfd->b.buf;
	    ed.b.len = gfd->b.len;
	    ed.data = NULL;
	    return (typ_msg(typ_dnsqname_msg,
			    TYP_EXTRACT,
			    &ed,
			    bs));
	  }
	return (-ERR_LAYER_NOSUCHFIELD);
      }
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	t_extract_data		ed;
	int			len;
	
	if (b->len < 1)
	  return (-ERR_LAYER_TRUNC);
	len = (t_u32)(t_u8)(b->buf[0]);
	if ((len & INDIR_MASK) == INDIR_MASK)
	  {
	    (*off) = 1 + 1;
	    return (0);
	  }
	ed.b.buf = b->buf;
	ed.b.len = b->len;
	ed.data = NULL;
	return (typ_msg(typ_dnsqname_msg,
			TYP_OFF,
			&ed,
			off));
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);
	int			len;

	if (b->len < 1)
	  return (-ERR_LAYER_TRUNC);
	len = (t_u32)(t_u8)(b->buf[0]);
	if ((len & INDIR_MASK) == INDIR_MASK)
	  {
	    (*mp) = lay_dnsqtype_msg;
	    return (0);
	  }
	if (len == 0)
	  (*mp) = &lay_dnsqtype_msg;
	else
	  (*mp) = &lay_dnsqname_msg;
	return (0);
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}

t_bit_field	dns_qr_bit_field =
{
  0,0
};

t_assoc		dns_qr_assocs[] =
{
  {"-",		(VOID_PTR)0},
  {"QR",	(VOID_PTR)1},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_qr_bfad =
{
  &dns_qr_bit_field,
  dns_qr_assocs
};

t_bit_field	dns_opcode_bit_field = 
{
  1,4
};

t_assoc		dns_opcode_assocs[] =
{
  {"query",		(VOID_PTR)0},
  {"inverse_query",	(VOID_PTR)1},
  {"server_status_request",(VOID_PTR)2},
  {NULL,		NULL},	
};

t_bit_field_assocs_data	dns_opcode_bfad =
{
  &dns_opcode_bit_field,
  dns_opcode_assocs
};

t_bit_field	dns_aa_bit_field = 
{
  5,5
};

t_assoc		dns_aa_assocs[] =
{
  {"-",		(VOID_PTR)0},
  {"AA",	(VOID_PTR)1},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_aa_bfad =
{
  &dns_aa_bit_field,
  dns_aa_assocs
};

t_bit_field	dns_tc_bit_field = 
{
  6,6
};

t_assoc		dns_tc_assocs[] =
{
  {"-",		(VOID_PTR)0},
  {"TC",	(VOID_PTR)1},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_tc_bfad =
{
  &dns_tc_bit_field,
  dns_tc_assocs
};

t_bit_field	dns_rd_bit_field = 
{
  7,7
};

t_assoc		dns_rd_assocs[] =
{
  {"-",		(VOID_PTR)0},
  {"RD",	(VOID_PTR)1},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_rd_bfad =
{
  &dns_rd_bit_field,
  dns_rd_assocs
};

t_bit_field	dns_ra_bit_field = 
{
  8,8
};

t_assoc		dns_ra_assocs[] =
{
  {"-",		(VOID_PTR)0},
  {"RA",	(VOID_PTR)1},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_ra_bfad =
{
  &dns_ra_bit_field,
  dns_ra_assocs
};

t_bit_field	dns_zero_bit_field = 
{
  9,11
};

t_bit_field	dns_rcode_bit_field = 
{
  11,15
};

t_assoc		dns_rcode_assocs[] =
{
  {"no_error",	(VOID_PTR)0},
  {"name_error",(VOID_PTR)3},
  {NULL,	NULL},
};

t_bit_field_assocs_data	dns_rcode_bfad =
{
  &dns_rcode_bit_field,
  dns_rcode_assocs
};

t_status	dns_sub(buf,len,sub_mp)
char		*buf;
int		len;
t_msg_proc	*sub_mp;
{
  t_dns		*dns;

  LAYER_DNS_CHECK(dns,buf,len);
  (*sub_mp) = &lay_dnsqname_msg;
  return (0);
}

t_field				dns_fields[] = 
{
  {"id",	OFFSET(t_dns *,id),	
   typ_nu16_msg,		NULL},
  {"qr",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_qr_bit_field)},
  {"Qr",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_qr_bfad)},
  {"opcode",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_opcode_bit_field)},
  {"Opcode",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_opcode_bfad)},
  {"aa",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_aa_bit_field)},
  {"Aa",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_aa_bfad)},
  {"tc",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_tc_bit_field)},
  {"Tc",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_tc_bfad)},
  {"rd",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_rd_bit_field)},
  {"Rd",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_rd_bfad)},
  {"ra",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_ra_bit_field)},
  {"Ra",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_ra_bfad)},
  {"zero",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_zero_bit_field)},
  {"rcode",	OFFSET(t_dns *,flags),
   typ_nu16bitfield_msg,	(VOID_PTR)(&dns_rcode_bit_field)},
  {"Rcode",	OFFSET(t_dns *,flags),
   typ_nu16bitfieldassoc_msg,	(VOID_PTR)(&dns_rcode_bfad)},
  {"nquestions",OFFSET(t_dns *,nquestions),
   typ_nu16_msg,		NULL},
  {"nanswers",	OFFSET(t_dns *,nanswers),
   typ_nu16_msg,		NULL},
  {"nauths",	OFFSET(t_dns *,nauths),
   typ_nu16_msg,		NULL},
  {"nadds",	OFFSET(t_dns *,nadds),
   typ_nu16_msg,		NULL},
  NULL_FIELD
};

char				*dns_itmpl = "\n\
<table _name=dns[%i%] width=100%%%% bgcolor=\"%%dnsColor%%\">\n\
<tr>\n\
<td align=center width=50%%%%><a href=\"set(dns[%i%].id)\">%%dns[%i%].id%%</a></td>\n\
<td align=center width=3%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Qr)\">%%dns[%i%].Qr%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=12%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Opcode)\">%%dns[%i%].Opcode%%</a>(<a href=\"set(dns[%i%].opcode)\">%%dns[%i%].opcode%%</a>)\n\
</_tiny>\n\
</td>\n\
<td align=center width=3%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Aa)\">%%dns[%i%].Aa%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=3%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Tc)\">%%dns[%i%].Tc%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=3%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Rd)\">%%dns[%i%].Rd%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=3%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Ra)\">%%dns[%i%].Ra%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=9%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].zero)\">%%dns[%i%].zero%%</a>\n\
</_tiny>\n\
</td>\n\
<td align=center width=12%%%%>\n\
<_tiny>\n\
<a href=\"set(dns[%i%].Rcode)\">%%dns[%i%].Rcode%%</a>(<a href=\"set(dns[%i%].rcode)\">%%dns[%i%].rcode%%</a>)\n\
</_tiny>\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=50%%%%><a href=\"set(dns[%i%].nquestions)\">%%dns[%i%].nquestions%%</a></td>\n\
<td align=center width=50%%%%><a href=\"set(dns[%i%].nanswers)\">%%dns[%i%].nanswers%%</a></td>\n\
</tr>\n\
<tr>\n\
<td align=center width=50%%%%><a href=\"set(dns[%i%].nauths)\">%%dns[%i%].nauths%%</a></td>\n\
<td align=center width=50%%%%><a href=\"set(dns[%i%].nadds)\">%%dns[%i%].nadds%%</a></td>\n\
</tr>\n\
</table>\n\
";

int				dns_nquestions_chan;

t_status			lay_dns_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_dns_msg,"dns");
      LAY_GET_FIELD_GENERIC(dns_fields);
      LAY_SET_FIELD_GENERIC(dns_fields);
      LAY_GET_FIELDS_GENERIC(dns_fields);
      LAY_GET_FIELD_TYP_GENERIC(dns_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_dns_msg,dns_itmpl);
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	
	(*off) = DNS_HLEN;
	return (0);
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);
	
	return (dns_sub(b->buf,b->len,mp));
      }
    case LAY_SUB_ID:
      {
	LAY_SUB_ID_ARGS(sid,mp);
	t_status		status;	
	t_dns			*dns;
	t_hash_elt		*he;
	t_buf			b;

	if (he = id_get(sid->id,
			&dns_nquestions_chan))
	  return (-ERR_LAYER_CHANINUSE);
	if (sid->b.len < DNS_HLEN)
	  return (-ERR_LAYER_TRUNC);
	dns = (t_dns *)(sid->b.buf);
	if ((status = id_add(sid->id,
			     &dns_nquestions_chan,
		 (VOID_PTR)((t_u32)(safe_ntohs(&dns->nquestions))))) < 0)
	  return (status);
	b.buf = sid->b.buf;
	b.len = sid->b.len;
	return (lay_msg(lay_dns_msg,
			LAY_SUB,
			&b,
			mp));
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}
