/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "lay_tcp.h"
#include "lay_ip.h"
#include "lay_data.h"
#include "typ_8.h"
#include "typ_16.h"
#include "typ_32.h"
#include "typ_port.h"
#include "typ_seqack.h"
#include "typ_time.h"

/*
Kind   Length   Meaning                           Reference
----   ------   -------------------------------   ---------
  0        -    End of Option List                 [RFC793]
  1        -    No-Operation                       [RFC793]
  2        4    Maximum Segment Lifetime           [RFC793]
  3        3    WSOPT - Window Scale              [RFC1323]
  4        2    SACK Permitted                    [RFC1072]
  5        N    SACK                              [RFC1072]
  6        6    Echo (obsoleted by option 8)      [RFC1072]
  7        6    Echo Reply (obsoleted by option 8)[RFC1072]
  8       10    TSOPT - Time Stamp Option         [RFC1323]
  9        2    Partial Order Connection Permitted[RFC1693]
 10        5    Partial Order Service Profile     [RFC1693]
 11             CC                                 [Braden]
 12             CC.NEW                             [Braden]
 13             CC.ECHO                            [Braden]
 14         3   TCP Alternate Checksum Request    [RFC1146]
 15         N   TCP Alternate Checksum Data       [RFC1146]
 16             Skeeter                           [Knowles]
 17             Bubba                             [Knowles]
 18         3   Trailer Checksum Option    [Subbu & Monroe]


TCP ALTERNATE CHECKSUM NUMBERS


Number  Description                     Reference
------- ------------------------------- ----------
   0    TCP Checksum                    [RFC-1146]
   1    8-bit Fletchers's algorithm     [RFC-1146]
   2    16-bit Fletchers's algorithm    [RFC-1146]
   3    Redundant Checksum Avoidance    [Kay]
*/

t_assoc				tcpopt_assocs[] = 
{
  {"eool",			(VOID_PTR)TCPOPT_EOOL},
  {"nop",			(VOID_PTR)TCPOPT_NOP},
  {"mss",			(VOID_PTR)TCPOPT_MSS},
  {"win",			(VOID_PTR)TCPOPT_WIN},
  {"ts",			(VOID_PTR)TCPOPT_TS},
  {NULL,			0},
};

t_field				tcpopt_simple_fields[] = 
{
  {"kind",	0,	typ_u8_msg,		NULL},
  {"Kind",	0,	typ_u8assoc_msg,	(VOID_PTR)tcpopt_assocs},
  NULL_FIELD
};

t_field				tcpopt_double_fields[] = 
{
  {"kind",	0,	typ_u8_msg,		NULL},
  {"Kind",	0,	typ_u8assoc_msg,	(VOID_PTR)tcpopt_assocs},
  {"len",	1,	typ_u8_msg,		NULL},
  NULL_FIELD
};

t_field				tcpopt_mss_fields[] = 
{
  {"kind",	0,	typ_u8_msg,		NULL},
  {"Kind",	0,	typ_u8assoc_msg,	(VOID_PTR)tcpopt_assocs},
  {"len",	1,	typ_u8_msg,		NULL},
  {"mss",	2,	typ_nu16_msg,		NULL},
  NULL_FIELD
};

t_field				tcpopt_win_fields[] = 
{
  {"kind",	0,	typ_u8_msg,		NULL},
  {"Kind",	0,	typ_u8assoc_msg,	(VOID_PTR)tcpopt_assocs},
  {"len",	1,	typ_u8_msg,		NULL},
  {"shift",	2,	typ_u8_msg,		NULL},
  NULL_FIELD
};

t_field				tcpopt_ts_fields[] = 
{
  {"kind",	0,	typ_u8_msg,		NULL},
  {"Kind",	0,	typ_u8assoc_msg,	(VOID_PTR)tcpopt_assocs},
  {"len",	1,	typ_u8_msg,		NULL},
  {"ts",	2,	typ_nu32_msg,		NULL},
  {"Ts",	2,	typ_time_msg,		NULL},
  {"tsreply",	6,	typ_nu32_msg,		NULL},
  {"Tsreply",	6,	typ_time_msg,		NULL},
  NULL_FIELD
};

t_field			*tcpopt_get_fields(buf,len,status)
char			*buf;
int			len;
t_status		*status;
{
  if (len < 1)
    {
      *status = -ERR_LAYER_TRUNC;
      return (NULL);
    }
  switch ((t_u8)(buf[0]))
    {
    case TCPOPT_EOOL:
    case TCPOPT_NOP:
      return (tcpopt_simple_fields);
    case TCPOPT_MSS:
      return (tcpopt_mss_fields);
    case TCPOPT_WIN:
      return (tcpopt_win_fields);
    case TCPOPT_TS:
      return (tcpopt_ts_fields);
    default:
      return (tcpopt_double_fields);
    }
}

char				*tcpopt_simple_itmpl = "\n\
<table _name=tcpopt[%i%] bgcolor=\"%%tcpoptColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Kind)\">%%tcpopt[%i%].Kind%%</a>\n\
(<a href=\"set(tcpopt[%i%].kind)\">%%tcpopt[%i%].kind%%</a>)\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

char				*tcpopt_double_itmpl = "\n\
<table _name=tcpopt[%i%] bgcolor=\"%%tcpoptColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=50%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Kind)\">%%tcpopt[%i%].Kind%%</a>\n\
(<a href=\"set(tcpopt[%i%].kind)\">%%tcpopt[%i%].kind%%</a>)\n\
</small>\n\
</td>\n\
<td align=center width=50%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].len)\">%%tcpopt[%i%].len%%</a>\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

char				*tcpopt_mss_itmpl = "\n\
<table _name=tcpopt[%i%] bgcolor=\"%%tcpoptColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Kind)\">%%tcpopt[%i%].Kind%%</a>\n\
(<a href=\"set(tcpopt[%i%].kind)\">%%tcpopt[%i%].kind%%</a>)\n\
</small>\n\
</td>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].len)\">%%tcpopt[%i%].len%%</a>\n\
</small>\n\
</td>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].mss)\">%%tcpopt[%i%].mss%%</a>\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

char				*tcpopt_win_itmpl = "\n\
<table _name=tcpopt[%i%] bgcolor=\"%%tcpoptColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Kind)\">%%tcpopt[%i%].Kind%%</a>\n\
(<a href=\"set(tcpopt[%i%].kind)\">%%tcpopt[%i%].kind%%</a>)\n\
</small>\n\
</td>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].len)\">%%tcpopt[%i%].len%%</a>\n\
</small>\n\
</td>\n\
<td align=center width=33%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].shift)\">%%tcpopt[%i%].shift%%</a>\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

char				*tcpopt_ts_itmpl = "\n\
<table _name=tcpopt[%i%] bgcolor=\"%%tcpoptColor%%\" width=100%%%%>\n\
<tr>\n\
<td align=center width=50%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Kind)\">%%tcpopt[%i%].Kind%%</a>\n\
(<a href=\"set(tcpopt[%i%].kind)\">%%tcpopt[%i%].kind%%</a>)\n\
</small>\n\
</td>\n\
<td align=center width=50%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].len)\">%%tcpopt[%i%].len%%</a>\n\
</small>\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Ts)\">%%tcpopt[%i%].Ts%%</a>(<a href=\"set(tcpopt[%i%].ts)\">%%tcpopt[%i%].ts%%</a>)\n\
</small>\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<small>\n\
<a href=\"set(tcpopt[%i%].Tsreply)\">%%tcpopt[%i%].Tsreply%%</a>\n\
(<a href=\"set(tcpopt[%i%].tsreply)\">%%tcpopt[%i%].tsreply%%</a>)\n\
</small>\n\
</td>\n\
</tr>\n\
</table>\n\
";

char			*tcpopt_get_itmpl(buf,len,status)
char			*buf;
int			len;
t_status		*status;
{
  if (len < 1)
    {
      *status = -ERR_LAYER_TRUNC;
      return (NULL);
    }
  switch ((t_u8)(buf[0]))
    {
    case TCPOPT_EOOL:
    case TCPOPT_NOP:
      return (tcpopt_simple_itmpl);
    case TCPOPT_MSS:
      return (tcpopt_mss_itmpl);
    case TCPOPT_WIN:
      return (tcpopt_win_itmpl);
    case TCPOPT_TS:
      return (tcpopt_ts_itmpl);
    default:
      return (tcpopt_double_itmpl);
    }
}

t_status		tcpopt_has_opt(buf,len,hsd)
char			*buf;
int			len;
t_has_opt_data		*hsd;
{
  if (len < 1)
    return (-ERR_LAYER_TRUNC);
  hsd->has_opt = FALSE;
  return (0);
}

char				*tcpoptdata_itmpl = "\n\
<table _name=tcpoptdata[%i%] width=100%%%% bgcolor=\"%%tcpoptColor%%\">\n\
<tr>\n\
<td width=100%%%%>\n\
<code>\n\
%%tcpoptdata[%i%].cookedhtmlbuf%%\n\
</code>\n\
</tr>\n\
</table>\n\
";

t_status			lay_tcpoptdata_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;	
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_tcpoptdata_msg,"tcpoptdata");
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_tcpoptdata_msg,tcpoptdata_itmpl);
    }
  return (lay_msg(lay_data_msg,
		  msg,
		  arg1,
		  arg2));
}

t_status			lay_tcpopt_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_tcpopt_msg,"tcpopt");
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);

	if (b->len < 1)
	  return (-ERR_LAYER_TRUNC);
	switch ((t_u8)(b->buf[0]))
	  {
	  case TCPOPT_EOOL:
	  case TCPOPT_NOP:
	    (*off) = 1;
	    return (0);
	  }
	if (b->len < 2)
	  return (-ERR_LAYER_TRUNC);
	(*off) = (t_off)(t_u8)(b->buf[1]);
	return (0);
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,sub_mp);
		
	if (b->len < 1)
	  return (-ERR_LAYER_TRUNC);
	switch ((t_u8)(b->buf[0]))
	  {
	  case TCPOPT_EOOL:
	    (*sub_mp) = &lay_tcpoptdata_msg;
	    return (0);
	  }
	(*sub_mp) = &lay_tcpopt_msg;
	return (0);
      }
    case LAY_GET_FIELD:
      {
	LAY_GET_FIELD_ARGS(gfd,bs);
	t_field			*fields;
	t_field			*field;
	
	if ((fields = tcpopt_get_fields(gfd->b.buf,
				       gfd->b.len,
				       &status)) == NULL)
	  return (status);
	field = NULL;
	while (fields->name)
	  {
	    if (!strcmp(fields->name,gfd->field))
	      field = fields;
	    fields++;
	  }
	if (!field)
	  return (-ERR_LAYER_NOSUCHFIELD);
	return (get_field_to_str(gfd->b.buf,
				 gfd->b.len,
				 field,
				 bs->str,
				 bs->max_len));
      }
    case LAY_SET_FIELD:
      {
	LAY_SET_FIELD_ARGS(gfd,str);
	t_field			*fields;
	t_field			*field;
	
	if ((fields = tcpopt_get_fields(gfd->b.buf,
				       gfd->b.len,
				       &status)) == NULL)
	  return (status);
	field = NULL;
	while (fields->name)
	  {
	    if (!strcmp(fields->name,gfd->field))
	      field = fields;
	    fields++;
	  }
	if (!field)
	  return (-ERR_LAYER_NOSUCHFIELD);
	return (set_field_from_str(gfd->b.buf,
				   gfd->b.len,
				   field,
				   str));
      }
    case LAY_GET_FIELD_TYP:
      {
	LAY_GET_FIELD_TYP_ARGS(gfd,gftd);
	t_field			*fields;	

	if ((fields = tcpopt_get_fields(gfd->b.buf,
				       gfd->b.len,
				       &status)) == NULL)
	  return (status);
	while (fields->name)
	  {
	    if (!strcmp(gfd->field,fields->name))
	      {
		gftd->mp = fields->mp;
		gftd->data = fields->data;
		return (0);
	      }
	    fields++;
	  }
	return (-ERR_LAYER_NOSUCHFIELD);
      }
    case LAY_GET_TMPL:
      {
	LAY_GET_TMPL_ARGS(gtd,bs);
	char		*itmpl;
	t_hash_elt	*he;
	int		i;

	if ((itmpl = tcpopt_get_itmpl(gtd->b.buf,
				   gtd->b.len,
				   &status)) == NULL)
	  return (status);
	if (he = id_get(gtd->id,
			(VOID_PTR)&lay_tcpopt_msg))
	  i = (int)(he->value);
	else
	  i = 0;
	if ((status = itmpl_format(i,
				   itmpl,
				   bs->str,
				   bs->max_len)) < 0)
	  return (status);
	if ((status = id_override(gtd->id,
				  (VOID_PTR)&lay_tcpopt_msg,
				  (VOID_PTR)(++i))) < 0)
	  return (status);
	return (0);
      }
    case LAY_HAS_OPT:
      {
	LAY_HAS_OPT_ARGS(b,hsd);
	
	return (tcpopt_has_opt(b->buf,b->len,hsd));
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}

t_mask_def		tcpflags_mask_defs[] = 
{
  {"F",			TCP_FIN},
  {"S",			TCP_SYN},
  {"R",			TCP_RST},
  {"P",			TCP_PUSH},
  {"A",			TCP_ACK},
  {"U",			TCP_URG},
  {NULL,		0},
};

t_mask_def		tcpflags_deluxe_mask_defs[] = 
{
  {"fin",		TCP_FIN},
  {"syn",		TCP_SYN},
  {"rst",		TCP_RST},
  {"push",		TCP_PUSH},
  {"ack",		TCP_ACK},
  {"urg",		TCP_URG},
  {NULL,		0}
};

t_bit_field		tcp_off_bit_field =
{
  0,3
};	

t_bit_field		tcp_x2_bit_field =
{
  4,7
};	

/* computes a tcp checksum.
   It is portable and deals with alignment.
   Returns checksum. */
t_u16			tcp_cksum(ip)
t_ip			*ip;
{
   t_tcp		*tcp;
   t_u16		*sptr;
   t_u16		len;
   unsigned long	sum;
   int			i;
   
   tcp = (t_tcp *)(((char *)ip) + ip_get_hl(ip) * 4);
   sum = 0;
   sptr = (t_u16 *)(&(ip->src));
   i = 0;
   while (i < IP_ALEN)
     {
       t_u16		u16;

       FBCOPY(sptr,&u16,sizeof (u16)); /* alignment */
       sum += UNSAFE_HTONS(u16);
       sptr++;
       i++;
     }
   sptr = (t_u16 *)tcp;
   len = UNSAFE_NTOHS(ip->len) - ip_get_hl(ip) * 4;
   sum += IP_PROTO_TCP + len;
   if (len % 2)
   {
      ((char *)tcp)[len] = 0;
      len += 1;
   }
   len >>= 1;
   i = 0;
   while (i < len)
     {
       t_u16		u16;

       FBCOPY(sptr,&u16,sizeof (u16)); /* alignment */
       sum += UNSAFE_HTONS(u16);
       sptr++;
       i++;
     }
   sum = (sum >> 16) + (sum & 0xffff);
   sum += (sum >> 16);
   return ((t_u16)(~sum & 0xffff));
}

VOID_FUNC		tcp_compute_sum(ip)
t_ip			*ip;
{
   t_tcp		*tcp;
   t_u16		zero;

   tcp = (t_tcp *)(((char *)ip) + ip_get_hl(ip) * 4);
   zero = 0;
   FBCOPY(&zero,&tcp->sum,sizeof (zero)); /* alignment */
   safe_htons(tcp_cksum(ip),&tcp->sum);
}

VOID_FUNC		tcp_set_proto(ip)
t_ip			*ip;
{
  ip->p = IP_PROTO_TCP;
}

VOID_FUNC		tcp_set_sport(tcp,sport)
t_tcp			*tcp;
int			sport;
{
  safe_htons(sport,&tcp->sport);
}

int			tcp_get_sport(tcp)
t_tcp			*tcp;
{
  return (safe_ntohs(&tcp->sport));
}	

VOID_FUNC		tcp_set_dport(tcp,dport)
t_tcp			*tcp;
int			dport;
{
  safe_htons(dport,&tcp->dport);
}

int			tcp_get_dport(tcp)
t_tcp			*tcp;
{
  return (safe_ntohs(&tcp->dport));
}	

VOID_FUNC		tcp_set_seq(tcp,seq)
t_tcp			*tcp;
t_tcp_seq		seq;
{
  safe_htonl(seq,&(tcp->seq));
}

t_tcp_seq		tcp_get_seq(tcp)
t_tcp			*tcp;
{
  return (safe_ntohl(&(tcp->seq)));
}

VOID_FUNC		tcp_set_ack(tcp,ack)
t_tcp			*tcp;
t_tcp_seq		ack;
{
  safe_htonl(ack,&(tcp->ack));
}

t_tcp_seq		tcp_get_ack(tcp)
t_tcp			*tcp;
{
  return (safe_ntohl(&(tcp->ack)));
}

VOID_FUNC		tcp_set_off(tcp,off)
t_tcp			*tcp;
int			off;
{
  bit_field_u8_set(&(tcp->offx2),
		   tcp_off_bit_field.from,
		   tcp_off_bit_field.to,
		   off);
}

int			tcp_get_off(tcp)
t_tcp			*tcp;
{
  t_u8			*offp;
  
  offp = bit_field_u8_get(&(tcp->offx2),
			  tcp_off_bit_field.from,
			  tcp_off_bit_field.to);
  return ((int)(*offp));
}

VOID_FUNC		tcp_set_x2(tcp,x2)
t_tcp			*tcp;
int			x2;
{
  bit_field_u8_set(&(tcp->offx2),
		   tcp_x2_bit_field.from,
		   tcp_x2_bit_field.to,
		   x2);
}

int			tcp_get_x2(tcp)
t_tcp			*tcp;
{
  t_u8			*x2p;
  
  x2p = bit_field_u8_get(&(tcp->offx2),
			 tcp_x2_bit_field.from,
			 tcp_x2_bit_field.to);
  return ((int)(*x2p));
}

VOID_FUNC		tcp_set_flags(tcp,flags)
t_tcp			*tcp;
int			flags;
{
  tcp->flags = (u_char)flags;
}

int			tcp_get_flags(tcp)
t_tcp			*tcp;
{
  return (tcp->flags);
}

VOID_FUNC		tcp_set_win(tcp,win)
t_tcp			*tcp;
int			win;
{
  safe_htons(win,&tcp->win);
}

int			tcp_get_win(tcp)
t_tcp			*tcp;
{
  return (safe_ntohs(&tcp->win));
}	

VOID_FUNC		tcp_set_sum(tcp,sum)
t_tcp			*tcp;
int			sum;
{
  safe_htons(sum,&tcp->sum);
}

int			tcp_get_sum(tcp)
t_tcp			*tcp;
{
  return (safe_ntohs(&tcp->sum));
}

VOID_FUNC		tcp_set_urp(tcp,urp)
t_tcp			*tcp;
int			urp;
{
  safe_htons(urp,&tcp->urp);
}

int			tcp_get_urp(tcp)
t_tcp			*tcp;
{
  return (safe_ntohs(&tcp->urp));
}	

t_status		tcp_off(buf,len,off)
char			*buf;
int			len;
t_off			*off;
{
  t_tcp			*tcp;

  LAYER_TCP_CHECK(tcp,buf,len);
  (*off) = tcp_get_off(tcp) * 4;
  return (0);
}

t_status		tcp_sub(buf,len,sub_mp)
char			*buf;
int			len;
t_msg_proc		*sub_mp;
{
  t_tcp			*tcp;

  LAYER_TCP_CHECK(tcp,buf,len);
  (*sub_mp) = &lay_data_msg;
  return (0);
}

t_status		tcp_sum(buf,len,up_buf,up_len)
char			*buf;
int			len;
char			*up_buf;
int			up_len;
{
  t_tcp			*tcp;
  t_ip			*ip;

  LAYER_TCP_CHECK(tcp,buf,len);
  LAYER_IP_CHECK(ip,up_buf,up_len);
  if (ip_get_len(ip) != up_len)
    return (-ERR_LAYER_BADLEN);
  if (ip_get_hl(ip) * 4 < IP_MINHLEN)
    return (-ERR_LAYER_BADHEADER);
  tcp_compute_sum(ip);
  return (0);
}

t_status		tcp_has_opt(buf,len,hsd)
char			*buf;
int			len;
t_has_opt_data		*hsd;
{
  t_tcp			*tcp;

  LAYER_TCP_CHECK(tcp,buf,len);
  if (tcp_get_off(tcp) * 4 > TCP_MINHLEN)
    {
      hsd->has_opt = TRUE;
      hsd->opt_off = TCP_MINHLEN;
      hsd->opt_len = tcp_get_off(tcp) * 4 - TCP_MINHLEN;
      hsd->opt_mp = lay_tcpopt_msg;
    }
  else
    {
      hsd->has_opt = FALSE;
    }
  return (0);
}

t_field				tcp_fields[] = 
{
  {"sport",	OFFSET(t_tcp *,sport),		typ_nu16_msg,		NULL},
  {"Sport",	OFFSET(t_tcp *,sport),		typ_port_resolved_msg,	"tcp"},
  {"dport",	OFFSET(t_tcp *,dport),		typ_nu16_msg,		NULL},
  {"Dport",	OFFSET(t_tcp *,dport),		typ_port_resolved_msg,	"tcp"},
  {"seq",	OFFSET(t_tcp *,seq),		typ_nu32_msg,		NULL},
  {"ack",	OFFSET(t_tcp *,ack),		typ_nu32_msg,		NULL},
  {"x2",	OFFSET(t_tcp *,offx2),
   typ_u8bitfield_msg,	(VOID_PTR)(&tcp_x2_bit_field)},
  {"off",	OFFSET(t_tcp *,offx2),
   typ_u8bitfield_msg,	(VOID_PTR)(&tcp_off_bit_field)},
  {"flags",	OFFSET(t_tcp *,flags),		typ_u8_msg,		NULL},
  {"Flags",	OFFSET(t_tcp *,flags),	 typ_u8mask_msg,  tcpflags_mask_defs},
  {"win",	OFFSET(t_tcp *,win),		typ_nu16_msg,		NULL},
  {"sum",	OFFSET(t_tcp *,sum),		typ_nu16_msg,		NULL},
  {"urp",	OFFSET(t_tcp *,urp),		typ_nu16_msg,		NULL},
  NULL_FIELD
};

char				*tcp_itmpl = "\n\
<table _name=tcp[%i%] width=100%%%% bgcolor=\"%%tcpColor%%\">\n\
<tr>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(tcp[%i%].Sport)\">%%tcp[%i%].Sport%%</a>\n\
(<a href=\"set(tcp[%i%].sport)\">%%tcp[%i%].sport%%</a>)\n\
</td>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(tcp[%i%].Dport)\">%%tcp[%i%].Dport%%</a>\n\
(<a href=\"set(tcp[%i%].dport)\">%%tcp[%i%].dport%%</a>)\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<a href=\"set(tcp[%i%].Seq)\">%%tcp[%i%].Seq%%</a>\n\
(<a href=\"set(tcp[%i%].seq)\">%%tcp[%i%].seq%%</a>)\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
<a href=\"set(tcp[%i%].Ack)\">%%tcp[%i%].Ack%%</a>\n\
(<a href=\"set(tcp[%i%].ack)\">%%tcp[%i%].ack%%</a>)\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=25%%%%>\n\
<a href=\"set(tcp[%i%].x2)\">%%tcp[%i%].x2%%</a>\n\
</td>\n\
<td align=center width=25%%%%>\n\
<a href=\"set(tcp[%i%].off)\">%%tcp[%i%].off%%</a>\n\
</td>\n\
<td align=center width=25%%%%>\n\
<a href=\"set(tcp[%i%].Flags)\">%%tcp[%i%].Flags%%</a>\n\
(<a href=\"set(tcp[%i%].flags)\">%%tcp[%i%].flags%%</a>)\n\
</td>\n\
<td align=center width=25%%%%>\n\
<a href=\"set(tcp[%i%].win)\">%%tcp[%i%].win%%</a>\n\
</td>\n\
</tr>\n\
<tr>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(tcp[%i%].sum)\">%%tcp[%i%].sum%%</a>\n\
</td>\n\
<td align=center width=50%%%%>\n\
<a href=\"set(tcp[%i%].urp)\">%%tcp[%i%].urp%%</a>\n\
</td>\n\
</tr>\n\
</table>\n\
";

t_status			lay_tcp_msg(msg,arg1,arg2)
t_msg				msg;
VOID_PTR			arg1;
VOID_PTR			arg2;
{
  t_status			status;

  switch (msg)
    {
      LAY_CLASS_GENERIC;
      LAY_NAME_ID_GENERIC((VOID_PTR)&lay_tcp_msg,"tcp");
      LAY_GET_FIELDS_GENERIC(tcp_fields);
      LAY_GET_FIELD_TYP_GENERIC(tcp_fields);
      LAY_GET_ITMPL_GENERIC((VOID_PTR)&lay_tcp_msg,tcp_itmpl);
    case LAY_GET_FIELD:
      {
	LAY_GET_FIELD_ARGS(gfd,bs);
	t_field			*field;
	t_field			*fields;

	field = NULL;
	fields = tcp_fields;
	while (fields->name)
	  {
	    if (!strcmp(fields->name,gfd->field))
	      field = fields;
	    fields++;
	  }
	if (!strcmp(gfd->field,"Seq") || !strcmp(gfd->field,"Ack"))
	  {
	    t_extract_data	ed;
	    
	    ed.b.buf = gfd->b.buf + TCP_SEQACK_OFF;
	    ed.b.len = gfd->b.len - TCP_SEQACK_OFF;
	    if (gfd->optional_id)
	      {
		t_seq_ack_data	sad;
		t_hash_elt	*he;
		t_tcp		*tcp;
		t_off		off;

		LAYER_TCP_CHECK(tcp,gfd->b.buf,gfd->b.len);
		if ((status = tcp_off(gfd->b.buf,
				      gfd->b.len,
				      &off)) < 0)
		  return (status);
		if (he = id_get(gfd->optional_id,
				&lay_ip_src_chan))
		  FBCOPY(&he->value,&sad.src,sizeof (sad.src));
		else
		  goto noipadvice;
		if (he = id_get(gfd->optional_id,
				&lay_ip_dst_chan))
		  FBCOPY(&he->value,&sad.dst,sizeof (sad.dst));
		else
		  goto noipadvice;
		sad.sport = safe_ntohs(&tcp->sport);
		sad.dport = safe_ntohs(&tcp->dport);
		sad.flags = tcp->flags;
		sad.datalen = gfd->b.len - off;
		sad.seq_wanted = !strcmp(gfd->field,"Seq");
		ed.data = &sad;
	      }
	    else
	      {
	      noipadvice:
		ed.data = NULL;
	      }
	    return (typ_msg(typ_seqack_msg,
			    TYP_EXTRACT,
			    &ed,
			    bs));
	    return (0);
	  }
	if (!field)
	  return (-ERR_LAYER_NOSUCHFIELD);
	return (get_field_to_str(gfd->b.buf,
				 gfd->b.len,
				 field,
				 bs->str,
				 bs->max_len));
      }
    case LAY_SET_FIELD:
      {
	LAY_SET_FIELD_ARGS(gfd,str);
	t_field			*field;
	t_field			*fields;
	
	field = NULL;
	fields = tcp_fields;
	while (fields->name)
	  {
	    if (!strcmp(fields->name,gfd->field))
	      field = fields; 
	    fields++;
	  }
	if (!strcmp(gfd->field,"Seq") || !strcmp(gfd->field,"Ack"))
	  return (-ERR_LAYER_NI);
	if (!field)
	  return (-ERR_LAYER_NOSUCHFIELD);
	return (set_field_from_str(gfd->b.buf,
				   gfd->b.len,
				   field,
				   str));
      }
    case LAY_OFF:
      {
	LAY_OFF_ARGS(b,off);
	
	return (tcp_off(b->buf,b->len,off));
      }
    case LAY_SUB:
      {
	LAY_SUB_ARGS(b,mp);

	return (tcp_sub(b->buf,b->len,mp));
      }
    case LAY_SUM:
      {
	LAY_SUM_ARGS(sd,unused);
	
	return (tcp_sum(sd->b.buf,
			sd->b.len,
			sd->up.buf,
			sd->up.len));
      }
    case LAY_HAS_OPT:
      {
	LAY_HAS_OPT_ARGS(b,hsd);
	
	return (tcp_has_opt(b->buf,b->len,hsd));
      }
    }
  return (-ERR_LAYER_NOLAYMETHOD);
}
