/*
 *
 */

#include <fcntl.h>
#include "lipfcache.h"
#include "lipdb.h"

extern t_boolean	verbose;	

t_dict			*lip_db_dict = NULL;

t_status		lip_db_init(VOID_DECL)
{
  t_status		status;

  if ((lip_db_dict = LIP_DICT_NEW(HASH_SMALL_BASE,
				VEC_BASE,
				&status)) == NULL)
    {
      err_print(-status,"LIP_DICT_NEW");
      return (status);
    }
  return (0);
}

t_status		lip_db_destroy_vec_str(he,unused)
t_hash_elt		*he;
VOID_PTR		unused;
{
  t_vec			*vec_str;

  vec_str = (t_vec *)(he->value);
  vec_str_delete(vec_str);
  return (0);
}

t_status		lip_db_destroy_dict(he,unused)
t_hash_elt		*he;
VOID_PTR		unused;
{
  t_dict		*dict;
  t_status		status;

  dict = (t_dict *)(he->value);
  status = dict_walk(dict,
		     (t_dict_walk_proc)lip_db_destroy_vec_str,
		     NULL);
  dict_delete(dict);
  return (status);
}

VOID_FUNC		lip_db_destroy(VOID_DECL)
{
  dict_walk(lip_db_dict,
	    (t_dict_walk_proc)lip_db_destroy_dict,
	    NULL);
  dict_delete(lip_db_dict);
}

t_dict			*lip_db_load(fname,status)
char			*fname;
t_status		*status;
{
  int			fd;
  t_parse_lines_context	plc;
  char			wordbuf[STR_BUFSIZ];
  t_dict		*dict;

  if (verbose)
    {
      fprintf(stderr,"loading db %s\n",fname);
    }
  if ((fd = open(fname,O_RDONLY)) < 0)
    {
      (*status) = -ERR_LIP_OPEN;
      err_print(-(*status),"opening db %s",fname);
      return (NULL);
    }
  if ((plc.vec = LIP_VEC_NEW(VEC_BASE,
			    status)) == NULL)
    {      
      close(fd);
      return (NULL);
    }
  plc.wordbuf = wordbuf;
  plc.wordbuflen = sizeof (wordbuf);
  plc.wordseps = " \t";
  plc.lineseps = "\n";
  parse_lines_init(&plc);
  while (1)
    {
      char		buf[BUFSIZ];
      int		cc;
      
      if ((cc = read(fd,buf,sizeof (buf))) < 0)
	{
	  (*status) = -ERR_LIP_READ;
	  break ;
	}
      if (cc == 0)
	{
	  (*status) = 0;
	  break ;
	}
      if (((*status) = parse_lines(&plc,buf,cc)) < 0)
	{
	  break ;
	}
    }
  if (((*status) = parse_lines_finnish(&plc)) < 0)
   goto bad;
  if ((dict = LIP_DICT_NEW(HASH_SMALL_BASE,
			  VEC_BASE,
			  status)) == NULL)
    return (NULL);
  VEC_FOR(plc.vec,t_vec *vec_str)
    {
      if (((*status) = dict_add(dict,VEC_AT(vec_str,0),vec_str)) < 0)
	{
	  dict_delete(dict);
	  goto bad;
	}
    }
  VEC_ENDFOR;
  if (((*status) = dict_add(lip_db_dict,fname,dict)) < 0)
    goto bad;
  vec_delete(plc.vec);
  close(fd);
  if (verbose)
    {
      fprintf(stderr,"successfully loaded db %s (%d records)\n",
	      fname,dict_count(dict));
    }
  return (dict);
bad:
  VEC_FOR(plc.vec,t_vec *vec_str)
    {
      vec_str_delete(vec_str);
    }
  VEC_ENDFOR;
  vec_delete(plc.vec);
  close(fd);
  return (NULL);  
}

t_dict			*lip_db_open(fname,status)
char			*fname;
t_status		*status;
{
  t_hash_elt		*he;

  if (he = dict_get(lip_db_dict,fname))
    return ((t_dict *)(he->value));
  else
    return (lip_db_load(fname,status));
}

char			*lip_db_get(fname,key,field,status)
char			*fname;
char			*key;
int			field;
t_status		*status;
{
  t_dict		*dict;
  t_hash_elt		*he;

  if ((dict = lip_db_open(fname,status)) == NULL)
    return (NULL);
  if (he = dict_get(dict,key))
    {
      t_vec		*vec;

      vec = (t_vec *)(he->value);
      if (field > (VEC_COUNT(vec) - 1))
	{
	  (*status) = -ERR_LIP_NOSUCHFIELD;
	  return (NULL);
	}
      else
	return (VEC_AT(vec,field));
    }
  (*status) = -ERR_LIP_NOSUCHKEY;
  return (NULL);
}
