/*
 *
 */

#include <fcntl.h>
#include "lipfcache.h"
#include "lipml.h"

extern t_boolean	verbose;

t_dict			*lip_ml_dict = NULL;

t_status		lip_ml_init(VOID_DECL)
{
  t_status		status;

  if ((lip_ml_dict = LIP_DICT_NEW(HASH_SMALL_BASE,
				VEC_BASE,
				&status)) == NULL)
    {
      err_print(-status,"LIP_DICT_NEW");
      return (status);
    }
  return (0);
}

t_status		lip_ml_destroy_dict(he,unused)
t_hash_elt		*he;
VOID_PTR		unused;
{
  t_vec			*vec;
  t_status		status;

  vec = (t_vec *)(he->value);
  VEC_FOR(vec,t_vec *vec_str)
    {
      vec_str_delete(vec_str);
    }
  VEC_ENDFOR;
  vec_delete(vec);
  return (status);
}

VOID_FUNC		lip_ml_destroy(VOID_DECL)
{
  dict_walk(lip_ml_dict,
	    (t_dict_walk_proc)lip_ml_destroy_dict,
	    NULL);
  dict_delete(lip_ml_dict);
}

t_vec			*lip_ml_load(fname,status)
char			*fname;
t_status		*status;
{
  int			fd;
  t_parse_lines_context	plc;
  char			wordbuf[STR_BUFSIZ];

  if (verbose)
    {
      fprintf(stderr,"loading db %s\n",fname);
    }
  if ((fd = open(fname,O_RDONLY)) < 0)
    {
      (*status) = -ERR_LIP_OPEN;
      err_print(-(*status),"opening db %s",fname);
      return (NULL);
    }
  if ((plc.vec = LIP_VEC_NEW(VEC_BASE,
			    status)) == NULL)
    {      
      close(fd);
      return (NULL);
    }
  plc.wordbuf = wordbuf;
  plc.wordbuflen = sizeof (wordbuf);
  plc.wordseps = " \t";
  plc.lineseps = "\n";
  parse_lines_init(&plc);
  while (1)
    {
      char		buf[BUFSIZ];
      int		cc;
      
      if ((cc = read(fd,buf,sizeof (buf))) < 0)
	{
	  (*status) = -ERR_LIP_READ;
	  break ;
	}
      if (cc == 0)
	{
	  (*status) = 0;
	  break ;
	}
      if (((*status) = parse_lines(&plc,buf,cc)) < 0)
	{
	  break ;
	}
    }
  if (((*status) = parse_lines_finnish(&plc)) < 0)
   goto bad;
  if (((*status) = dict_add(lip_ml_dict,fname,plc.vec)) < 0)
    goto bad;
  close(fd);
  if (verbose)
    {
      fprintf(stderr,"successfully loaded ml %s (%d records)\n",
	      fname,VEC_COUNT(plc.vec));
    }
  return (plc.vec);
bad:
  VEC_FOR(plc.vec,t_vec *vec_str)
    {
      vec_str_delete(vec_str);
    }
  VEC_ENDFOR;
  vec_delete(plc.vec);
  close(fd);
  return (NULL);  
}

t_vec			*lip_ml_open(fname,status)
char			*fname;
t_status		*status;
{
  t_hash_elt		*he;

  if (he = dict_get(lip_ml_dict,fname))
    return ((t_vec *)(he->value));
  else
    return (lip_ml_load(fname,status));
}

char			*lip_ml_get(fname,key,field,status)
char			*fname;
char			*key;
int			field;
t_status		*status;
{
  t_vec			*vec;

  if ((vec = lip_ml_open(fname,status)) == NULL)
    return (NULL);
  VEC_FOR(vec,t_vec *vec_str)
    {
      if (gmatch(key,VEC_AT(vec_str,0)))
	{
	  if (field > (VEC_COUNT(vec_str) - 1))
	    {
	      (*status) = -ERR_LIP_NOSUCHFIELD;
	      return (NULL);
	    }
	  else
	    return (VEC_AT(vec_str,field));
	}
    }
  VEC_ENDFOR;
  (*status) = -ERR_LIP_NOMATCH;
  return (NULL);
}
