/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include <sys/types.h>
#include <sys/uio.h>
#include <pcap.h>
#include <X11/Intrinsic.h>
#include "xip.h"
#include "XmgFancy.h"
#include "pkt.h"
#include "lay_ether.h"
#include "lay_null.h"
#include "lay_data.h"

extern XtAppContext		app_context;
extern Widget			toplevel;
extern Widget			pktbox;
extern t_boolean		capflow;
extern t_boolean		big_packets;
extern t_boolean		pcap_fork;
extern char			*pcap_dev;
extern char			hard_filter[];
extern int			pcap_snaplen;
extern t_boolean		pcap_promisc;
extern int			pcap_to_ms;
extern t_boolean		pcap_optimize;
extern char			*pcap_fname;
extern int			pcap_cnt;
extern t_boolean		verbose;

pcap_t				*pcap;
int				packets_treated = 0;
int				packets_dropped = 0;
pid_t				pcap_pid;

char				*pcap_not_implemented_tmpl = "\n\
<!--pcap_tmpl-->\n\
<table width=100%%%%>\n\
<tr>\n\
<td align=center width=100%%%%>\n\
Not implemented\n\
</tr>\n\
</table>\n\
";

#define NOTREALLYIMPLEMENTED(MsgProc,Name,Offset) \
t_status		MsgProc(msg,arg1,arg2)\
t_msg			msg;\
VOID_PTR		arg1;\
VOID_PTR		arg2;\
{\
  switch (msg)\
    {\
       LAY_CLASS_GENERIC;\
       LAY_NAME_GENERIC(Name);\
       LAY_GET_TMPL_GENERIC(pcap_not_implemented_tmpl);\
     case LAY_SUB:\
       {\
	 LAY_SUB_ARGS(b,mp);\
\
	 (*mp) = lay_data_msg;\
	 return (0);\
       }\
    case LAY_OFF:\
      {\
	LAY_OFF_ARGS(b,off);\
\
	(*off) = (Offset);\
	return (0);\
      }\
    }\
  return (-ERR_LAYER_NOLAYMETHOD);\
}

NOTREALLYIMPLEMENTED(lay_slip_msg,"slip",16)
NOTREALLYIMPLEMENTED(lay_ppp_msg,"ppp",4)
#ifdef PCAP_FDDIPAD
NOTREALLYIMPLEMENTED(lay_fddi_msg,"fddi",21 + PCAP_FDDIPAD)
#else
NOTREALLYIMPLEMENTED(lay_fddi_msg,"fddi",21)
#endif
NOTREALLYIMPLEMENTED(lay_atm_rfc1483_msg,"atm_rfc1483",8)

/* compiles the bpf program and sets it to the main pcap structure. 
   It exits(2) if something goes wrong. */
VOID_FUNC			set_hard_filter(pcap,hard_filter)
pcap_t				*pcap;
char				*hard_filter;
{
  struct bpf_program		bp;

  if (pcap_compile(pcap,
                   &bp,
                   hard_filter,
                   pcap_optimize,
                   0) < 0)
    {
      fprintf(stderr,"problems in pcap_compile: %s\n",hard_filter);
      exit(1);
    }
  if (pcap_setfilter(pcap,&bp) < 0)
    {
      fprintf(stderr,"can't set pcap_setfilter\n");
      exit(1);
    }
}

struct				s_singleton
{
  struct pcap_pkthdr		*pp;
  char				*buf;
};

/* transforms a pcap "link type" to a "lay" msg_proc.
   Returns 0 if OK, -ERR_XIP_NOENT if no equivalent is found. */
t_status			lay_get_from_pcap(pcap_link_type,mp)
int				pcap_link_type;
t_msg_proc			*mp;		/* "Lay" msg_proc */
{
  switch (pcap_link_type)
    {
    case DLT_EN10MB:
      (*mp) = &lay_ether_msg;
      return (0);
    case DLT_SLIP:
      (*mp) = &lay_slip_msg;
      return (0);
    case DLT_NULL:
      (*mp) = &lay_null_msg;
      return (0);
    case DLT_PPP:
      (*mp) = &lay_ppp_msg;
      return (0);
    case DLT_FDDI:
      (*mp) = &lay_fddi_msg;
      return (0);
    case DLT_IEEE802:
      (*mp) = &lay_ether_msg;
      return (0);
#ifdef DLT_ATM_RFC1483
    case DLT_ATM_RFC1483:
      (*mp) = &lay_atm_rfc1483_msg;
      return (0);
#endif
    }
  return (-ERR_XIP_NOENT);
}

/* fills out a "pkt" structure against pcap information.
   then create graphical packet. It uses lay_get_from_pcap(3).
   Returns 0 if OK, Might return various errors. */
VOID_FUNC			make_packet_from_pp_and_buf(pp,buf)
struct pcap_pkthdr		*pp;
char				*buf;
{
  t_status			status;
  t_pkt				pkt;
  t_pkt				*npkt;

  if (capflow)
    {
      if ((status = lay_get_from_pcap(pcap_datalink(pcap),&pkt.mp)) < 0)
	pkt.mp = &lay_data_msg;
      pkt.buf = buf;
      pkt.len = pp->caplen;
      pkt.netlen = pp->len;
      FBCOPY(&pp->ts,&pkt.ts,sizeof (struct timeval));
      if (verbose)
	{
	  char			text[STR_BUFSIZ];

	  text[0] = 0;
	  if ((status = xip_shortcuts_html_do(&pkt,
					      text,
					      sizeof (text))) < 0)
	    ;
	  fprintf(stderr,"%s\n",text);
	}
      if ((npkt = pkt_dup(&pkt,
			  XIP_ALLOC_PROC,
			  XIP_FREE_PROC,
			  &status)) == NULL)
	{
	  XmgErrPrint(toplevel,-status,"pkt_dup");
	  return ;
	}
      XipCreatePkt(big_packets?toplevel:pktbox,
		   npkt,
		   TRUE);
      packets_treated++;
    }
  else
    {
      packets_dropped++;
    }
}

/* is a pcap_handler. 
   calls make_packet_from_pp_and_buf(3). */
VOID_FUNC			read_pcap_handler(s,pp,buf)
struct s_singleton		*s;
struct pcap_pkthdr		*pp;
char				*buf;
{
  make_packet_from_pp_and_buf(pp,buf);
}

/* is a XtInputCallbackProc.
   It reads pcap file descriptor. */
VOID_FUNC			read_pcap(closure,fd,id)
XtPointer			closure;
int				*fd;
XtInputId			*id;
{
  struct s_singleton		s;
  
  if (pcap_read(pcap,
		1,
		(pcap_handler)read_pcap_handler,
		(u_char *)(&s)) < 0)
    {
    }
}

/* is a XtInputCallbackProc.
   It reads pipe file descriptor. */
VOID_FUNC			read_pipe(closure,fd,id)
XtPointer			closure;
int				*fd;
XtInputId			*id;
{
  int				cc;
  struct pcap_pkthdr		pp;
  char				buf[BUFSIZ];
  
  if ((cc = read(*fd,&pp,sizeof (pp))) < 0)
    {
      err_print(ERR_XIP_READ,"reading pp");
      return ;
    }
  assert(pp.len < sizeof (buf));
  if ((cc = read(*fd,buf,pp.len)) < 0)
    {
      err_print(ERR_XIP_READ,"reading buf");
      return ;
    }
  make_packet_from_pp_and_buf(&pp,buf);
}

/* initializes the pcap part of xip.
   /* opens pcap, possibly opening "offline" if pcap_fname is not NULL.
   If pcap_dev is NULL, it performs pcap_lookupdev(3).
   It needs "app_context" to be initialized since it uses XtAppAddInput(3).
   It pcap_fname is not NULL, it loops on the file. */
t_status			xip_cap_init(VOID_DECL)
{
  char				errbuf[BUFSIZ];

  if (pcap_fname)
    {
      if ((pcap = pcap_open_offline(pcap_fname,
				    errbuf)) == NULL)
	{
	  fprintf(stderr,"pcap_open_offline: %s\n",errbuf);
	  exit(1);
	}
    }
  else
    {
      if (!pcap_dev)
	if ((pcap_dev = pcap_lookupdev(errbuf)) == NULL)
	  {
	    fprintf(stderr,"%s\n",errbuf);
	    exit(1);
	  }
      if ((pcap = pcap_open_live(pcap_dev,
				 pcap_snaplen,
				 pcap_promisc,
				 pcap_to_ms,
				 errbuf)) == NULL)
	{
	  fprintf(stderr,"pcap_open_live: %s\n",errbuf);
	  exit(1);
	}
    }
  set_hard_filter(pcap,hard_filter);
  if (pcap_fname)
    {
      pcap_loop(pcap,
		pcap_cnt,
		(pcap_handler)read_pcap_handler,
		NULL);
    }
  else
    {
      if (pcap_fork)
	{
	  int		p[2];

	  if (pipe(p) < 0)
	    {
	      err_print(ERR_XIP_SYSCALL,"pipe");
	      exit(1);
	    }
	  if ((pcap_pid = fork()) < 0)
	    {
	      err_print(ERR_XIP_SYSCALL,"fork");
	      exit(1);
	    }
	  if (pcap_pid)
	    {
	      close(p[1]);
	      XtAppAddInput(app_context,
			    p[0],
			    (XtPointer)XtInputReadMask,
			    read_pipe,
			    NULL);
	    }
	  else
	    {
	      close(p[0]);
	      while (1)
		{
		  struct pcap_pkthdr	pp;
		  char			*buf;

		  if (buf = (char *)pcap_next(pcap,&pp))
		    {
		      struct iovec	iov[2];
		      
		      iov[0].iov_base = (caddr_t)(&pp);
		      iov[0].iov_len = sizeof (pp);
		      iov[1].iov_base = buf;
		      iov[1].iov_len = pp.len;
		      if (writev(p[1],iov,2) < 0)
			{
			  err_print(ERR_XIP_WRITE,"writev");
			}
		    }
		}
	    }
	}
      else
	{
	  XtAppAddInput(app_context,
			pcap_fileno(pcap),
			(XtPointer)XtInputReadMask,
			read_pcap,
			NULL);
	}
    }
  return (0);
}
