/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <fcntl.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include "XmgMacro.h"
#include "XmgFancy.h"
#include "Xmg/Group.h"
#include "Tabs.h"
#include "TextField.h"
#include "lay_data.h"
#include "lay_ip.h"
#include "lay_udp.h"
#include "lay_pad.h"
#include "typ_lay.h"
#include "typ_ethaddr.h"
#include "typ_boolean.h"
#include "layer.h"
#include "nbgethost.h"
#include "xip.h"
#include "xipvar.h"
#ifdef HAVE_DLOPEN
# include <dlfcn.h>
#endif

extern Widget		toplevel;
extern t_boolean	verbose;
extern Pixmap		xipicon_pixmap;
extern char		xiprc[];
extern t_u32		xip_do_opts;

t_u32			tmpl_bufsiz = 16 * BUFSIZ;
t_u32			html_bufsiz = 16 * BUFSIZ;
t_u32			htmlize_bufsiz = 16 * BUFSIZ;
struct in_addr		ghost_ip = {-2};
t_ether_addr		ghost_ether = {0x42,0x42,0x42,0x42,0x42,0x43};
t_u16			ip_id = 1;

char			*tmpl_buf = NULL;
char			*html_buf = NULL;
char			*htmlize_buf = NULL;

/* is a t_msg_proc.
   manages memory zones. Ptr is passed as ed->data and size is passed as
   an t_u32 in ed->b.buf and ed->b.len. */
t_status		typ_memoryzone_msg(msg,arg1,arg2)
t_msg			msg;
VOID_PTR		arg1;
VOID_PTR		arg2;
{
  t_status		status;

  switch (msg)
    {
      TYP_CLASS_GENERIC;
      TYP_NAME_GENERIC("memoryzone");
    case TYP_EXTRACT:
      {
	TYP_EXTRACT_ARGS(ed,bs);
	
	return (typ_msg(typ_u32_msg,
			msg,
			arg1,
			arg2));
      }
    case TYP_INSERT:
      {
	TYP_INSERT_ARGS(ed,str);
	char		**zone;
	t_u32		siz;	

	if ((status = typ_msg(typ_u32_msg,
			      msg,
			      arg1,
			      arg2)) < 0)
	  return (status);
	assert(ed->b.len == sizeof (siz));
	siz = *((t_u32 *)(ed->b.buf));
	zone = (char **)(ed->data);
	if (*zone)
	  {
	    if (((*zone) = XIP_REALLOC_PROC(*zone,
					    siz,
					    "xipcf",
					    "typ_memoryzone_msg:ptr",
					    &status)) == NULL)
	      {
		err_print(-status,"XIP_REALLOC_PROC");
		exit(1);
	      }
	  }
	else
	  {
	    if (((*zone) = XIP_ALLOC_PROC(siz,
					  "xipcf",
					  "typ_memoryzone_msg:ptr",
					  &status)) == NULL)
	      {
		err_print(-status,"XIP_ALLOC_PROC");
		exit(1);
	      }
	  }
	return (0);
      }
    }
  return (-ERR_XIP_NOMETHOD);
}

#ifdef HAVE_DLOPEN
/* is a t_msg_proc.
   Manages plugins. It detects new plugins inserted. */
t_status		typ_plugins_msg(msg,arg1,arg2)
t_msg			msg;
VOID_PTR		arg1;
VOID_PTR		arg2;
{
  t_status		status;
  static t_vec		*plugins = NULL;

  if (!plugins)
    {
      if ((plugins = XIP_VEC_NEW(VEC_BASE,
				 &status)) == NULL)
	return (status);
    }
  switch (msg)
    {
      TYP_CLASS_GENERIC;
      TYP_NAME_GENERIC("plugins");
    case TYP_EXTRACT:
      {
	TYP_EXTRACT_ARGS(ed,bs);

	VEC_FOR(plugins,char *path)
	  {
	    if ((status = str_cat_str(bs->str,
				      bs->max_len,
				      path)) < 0)
	      return (status);
	    if (VEC_IDX != (VEC_COUNT(plugins) - 1))
	      if ((status = str_cat_char(bs->str,
					 bs->max_len,
					 ';')) < 0)
		return (status);
	  }
	VEC_ENDFOR;
	return (0);
      }
    case TYP_INSERT:
      {
	TYP_INSERT_ARGS(ed,str);
	t_vec		*vec;

	if ((vec = XIP_VEC_NEW(VEC_BASE,
			       &status)) == NULL)
	  return (status);
	if ((status = vec_str_split(vec,
				    str,
				    ';')) < 0)
	  {
	    vec_str_delete(vec);
	    return (status);
	  }
	VEC_FOR(vec,char *path)
	  {
	    VOID_PTR		plugin;
	    t_xip_plugin_proc	xip_plugin_proc;
	    
	    if (vec_str_index(plugins,path) >= 0)
	      {
		continue ;
	      }
	    if ((plugin = dlopen(path,RTLD_NOW)) == NULL)
	      {
		err_print(ERR_XIP_SUBROUTINE,"%s: %s",dlerror());
	      }
	    else
	      {
		if ((xip_plugin_proc =
		     (t_xip_plugin_proc)dlsym(plugin,"xip_plugin_init"))
		    == NULL)
		  {
		    err_print(ERR_XIP_SUBROUTINE,"%s: %s",dlerror());
		    dlclose(plugin);
		  }
		else
		  {
		    if ((status = xip_plugin_proc()) < 0)
		      return (status);
		    if ((status = vec_str_add(plugins,path)) < 0)
		      return (status);
		  }
	      }
	  }
	VEC_ENDFOR;
	vec_str_delete(vec);
	return (0);
      }
    }
  return (-ERR_XIP_NOMETHOD);
}
#endif

t_xip_var		xip_vars[] = 
{
#define NOT_A_REAL_OFFSET	((t_off)(NULL))
#define NOT_A_REAL_SIZE		0

  {"tmpl_bufsiz",
   (t_off)(&tmpl_bufsiz),
   sizeof (tmpl_bufsiz),
   typ_memoryzone_msg,
   &tmpl_buf},
  {"html_bufsiz",
   (t_off)(&html_bufsiz),
   sizeof (html_bufsiz),
   typ_memoryzone_msg,
   &html_buf},
  {"htmlize_bufsiz",
   (t_off)(&htmlize_bufsiz),
   sizeof (htmlize_bufsiz),
   typ_memoryzone_msg,	
   &htmlize_buf},
  {"lay_data_bytes_per_line",
   (t_off)(&lay_data_bytes_per_line),
   sizeof (lay_data_bytes_per_line),
   typ_u32_msg,
   NULL},
  {"lay_udp_perform_sum",
   (t_off)(&lay_udp_perform_sum),
   sizeof (lay_udp_perform_sum),
   typ_boolean_msg,
   NULL},
  {"lay_udp_guess_mode",
   (t_off)(&lay_udp_guess_mode),
   sizeof (lay_udp_guess_mode),
   typ_boolean_msg,
   NULL},
  {"layer_verb_level",
   (t_off)(&layer_verb_level),
   sizeof (layer_verb_level),
   typ_u32mask_msg,
   layer_verb_level_mask_defs},
  {"layer_base",(t_off)(&layer_base),
   sizeof (layer_base),
   typ_u32_msg,
   NULL},
  {"ghost_ip",
   (t_off)(&ghost_ip),
   sizeof (ghost_ip),
   typ_inaddr_resolved_msg,
   NULL},
  {"ghost_ether",
   (t_off)(&ghost_ether),
   sizeof (ghost_ether),
   typ_etheraddr_msg,
   NULL},
  {"ip_id",
   (t_off)(&ip_id),
   sizeof (ip_id),
   typ_u16_msg,
   NULL},
  {"nbgethost_timeout",
   (t_off)(&nbgethost_timeout),
   sizeof (nbgethost_timeout),
   typ_u32_msg,
   NULL},
  {"lay_pad_nbytes",
   (t_off)(&lay_pad_nbytes),
   sizeof (lay_pad_nbytes),
   typ_u32_msg,
   NULL},
  {"lay_pad_sub_lay",
   (t_off)(&lay_pad_sub_lay),
   sizeof (lay_pad_sub_lay),
   typ_lay_msg,
   NULL},
  {"etheraddr_dec_way",
   (t_off)(&etheraddr_dec_way),
   sizeof (etheraddr_dec_way),
   typ_boolean_msg,
   NULL},
  {"xip_do_opts",
   (t_off)(&xip_do_opts),
   sizeof (xip_do_opts),
   typ_boolean_msg,
   NULL},
#ifdef HAVE_DLOPEN
  {"plugins",
   NOT_A_REAL_OFFSET,
   NOT_A_REAL_SIZE,
   typ_plugins_msg,
   NULL},
#endif
  NULL_XIP_VAR
};

t_vec			*xip_vars_vec = NULL;

t_status		xip_var_add(xv)
t_xip_var		*xv;
{
  t_xip_var		*own_xv;
  t_status		status;
  
  if ((own_xv = XIP_ALLOC_PROC(sizeof (t_xip_var),
			       "xip",
			       "xip_pkt_bar_add:ptr",
			       &status)) == NULL)
    return (status);
  FBCOPY(xv,own_xv,sizeof (t_xip_var));
  if ((status = vec_add(xip_vars_vec,
			own_xv)) < 0)
    {
      XIP_FREE_PROC(own_xv,
		    "xip",
		    "*:ptr");
      return (status);
    }
  return (0);
}

/* inits vars context.
   Currently, it allocates html_buf, tmpl_buf and htmlize_buf.
   Returns 0 if OK. Might return various errors */
t_status		xip_vars_init(VOID_DECL)
{
  t_status		status;
  t_xip_var		*xv;

  if ((xip_vars_vec = XIP_VEC_NEW(VEC_ONE_BASE,
				 &status)) == NULL)
    return (status);
  xv = xip_vars;
  while (xv->name)
    {
      if ((status = xip_var_add(xv)) < 0)
	return (status);
      xv++;
    }
  if ((html_buf = XIP_ALLOC_PROC(html_bufsiz,
				 "xipcf",
				 "xip_vars_init:ptr",
				 &status)) == NULL)
    return (status);
  if ((tmpl_buf = XIP_ALLOC_PROC(tmpl_bufsiz,
				 "xipcf",
				 "xip_vars_init:ptr",
				 &status)) == NULL)
    return (status);
  if ((htmlize_buf = XIP_ALLOC_PROC(htmlize_bufsiz,
				    "xipcf",
				    "xip_vars_init:ptr",
				    &status)) == NULL)
    return (status);
  return (0);
}

/* destroys vars context.
   Currently, it destroys html_buf, tmpl_buf and htmlize_buf. */
VOID_FUNC		xip_vars_destroy(VOID_DECL)
{
  vec_ptr_delete(xip_vars_vec);
  XIP_FREE_PROC(html_buf,
		"xipcf",
		"*:ptr");
  XIP_FREE_PROC(tmpl_buf,
		"xipcf",
		"*:ptr");
  XIP_FREE_PROC(htmlize_buf,
		"xipcf",
		"*:ptr");
}

/* gets t_xip_var structure associated with name.
   Returns NULL of not found. */
t_xip_var		*xip_var_get_from_name(name)
char			*name;
{
  VEC_FOR(xip_vars_vec,t_xip_var *xv)
    {
      if (!strcmp(name,xv->name))
	return (xv);
    }
  VEC_ENDFOR;
  return (NULL);
}

/* sets a configuration variable.
   Note: it prints an error message and returns 0 if var doesn't exist.
   Returns 0 if OK. Might return various errors. */
t_status		xip_var_set(name,value)
char			*name;
char			*value;
{
  t_status		status;
  t_extract_data	ed;
  t_xip_var		*xv;

  if (verbose)
    fprintf(stderr,"setting %s to %s\n",name,value);
  if (!(xv = xip_var_get_from_name(name)))
    {
      err_print(ERR_XIP_NOSUCHVAR,"no such var %s",name);
      return (0);
    }
  ed.b.buf = (char *)(xv->offset);
  ed.b.len = xv->size;
  ed.data = xv->data;
  return (typ_msg(xv->mp,
		  TYP_INSERT,
		  &ed,
		  value));
}

/* gets a variable value to a string.
   Returns 0 if OK. -ERR_XIP_NOSUCHVAR if var doesn't exist. Might
   return various errors. */
t_status		xip_var_get(name,str,max_len)
char			*name;
char			*str;
int			max_len;
{
  t_status		status;
  t_extract_data	ed;
  t_xip_var		*xv;
  t_bridled_str		bs;

  if (!(xv = xip_var_get_from_name(name)))
    return (-ERR_XIP_NOSUCHVAR);
  ed.b.buf = (char *)(xv->offset);
  ed.b.len = xv->size;
  ed.data = xv->data;
  bs.str = str;
  bs.max_len = max_len;
  return (typ_msg(xv->mp,
		  TYP_EXTRACT,
		  &ed,
		  &bs));
}

/* saves configuration to a file.
   Returns 0 if OK. -ERR_XIP_OPEN if fopen(3) fails. Might return various
   errors as it calls typ methods. */
t_status		xip_cf_save(fname)
char			*fname;
{
  FILE			*f;
  t_status		status;

  if ((f = fopen(fname,"w+")) == NULL)
    return (-ERR_XIP_OPEN);
  fprintf(f,"#XIPCF\n");
  VEC_FOR(xip_vars_vec,t_xip_var *xv)
    {
      char		value[STR_BUFSIZ];
      t_extract_data	ed;
      t_bridled_str	bs;
      
      value[0] = 0;
      ed.b.buf = (char *)(xv->offset);
      ed.b.len = xv->size;
      ed.data = xv->data;
      bs.str = value;
      bs.max_len = sizeof (value);
      if ((status = typ_msg(xv->mp,
			    TYP_EXTRACT,
			    &ed,
			    &bs)) < 0)
	{
	  fclose(f);
	  return (status);
	}
      fprintf(f,"%s = %s\n",xv->name,value);
    }
  VEC_ENDFOR;
  fclose(f);
  return (status);
}

/* is a t_vec_cmp_proc.
   Sorts vars with "plugins" first. */
int			xip_cf_load_cmp(p1,p2)
VOID_PTR		*p1;
VOID_PTR		*p2;
{
  t_hash_elt		*he1;
  t_hash_elt		*he2;
  
  he1 = (t_hash_elt *)(*p1);
  he2 = (t_hash_elt *)(*p2);
  if (!strcmp(he1->key,"plugins"))
    {
      if (!strcmp(he2->key,"plugins"))
	return (0);
      else
	return (-1);
    }
  else
    {
      if (!strcmp(he2->key,"plugins"))
	return (1);
      else
	return (0);
    }
}

/* is a t_dict_walk_proc.
   Assign value to key in using xip_var_set(3).
   If global verbose is TRUE, then it prints a shortcut of the operation.
   Anyway this function will print an error message if there is a problem
   with xip_var_set(3).
   Returns 0 if OK. Might return various errors. */
t_status		xip_cf_load_walk(he,unused)
t_hash_elt		*he;
VOID_PTR		unused;
{
  t_status		status;
   
  if ((status = xip_var_set(he->key,he->value)) < 0)
    err_print(-status,"xip_var_set %s",he->key);
  return (status);
}

/* loads a configuration file.
   Note: it might print various error and verbose messages as it calls
   xip_cf_load_walk(3).
   Returns 0 if OK. -ERR_XIP_OPEN if open(2) fails.
   Might return various errors. */
t_status		xip_cf_load(fname)
char			*fname;
{
  int			fd;
  t_status		status;
  t_dict		*vars;
  t_hash_elt		*he;

  if ((fd = open(fname,O_RDONLY)) < 0)
    return (-ERR_XIP_OPEN);
  vars = NULL;
  if ((vars = XIP_DICT_NEW(HASH_SMALL_BASE,
			   VEC_BASE,
			   &status)) == NULL)
    goto end;
  if ((status = parse_vars(fd,
			   '=',
			   vars,
			   dict_str_override)) < 0)
    goto end;
  if ((status = dict_walk_sorted_cmp(vars,
				     (t_dict_walk_proc)xip_cf_load_walk,
				     (t_vec_cmp_proc)xip_cf_load_cmp,
				     NULL)) < 0)
    goto end;
  status = 0;
end:
  close(fd);
  if (vars)
    dict_str_delete(vars);
  return (status);
}

VOID_FUNC		XipCfCancelCommand(w,shell,cbs)
Widget			w;
Widget			shell;
VOID_PTR		cbs;
{
  XtDestroyWidget(shell);
}

char			*xip_var_get_static(var)
char			*var;
{
  t_status		status;
  static char		str[BUFSIZ];

  str[0] = 0;
  if ((status = xip_var_get(var,str,sizeof (str))) < 0)
    {
      err_print(-status,"xip_var_get %s",var);
    }
  return (str);
}

VOID_FUNC		XipCfOkCommand(w,shell,cbs)
Widget			w;
Widget			shell;
VOID_PTR		cbs;
{
  t_status		status;

  VEC_FOR(xip_vars_vec,t_xip_var *xv)
    {
      Widget		w;
      char		fqvarname[STR_BUFSIZ];

      fqvarname[0] = 0;
      if ((status = str_cat_fmt_va(fqvarname,
				   sizeof (fqvarname),
				   "*%s",
				   xv->name)) < 0)
	{
	  err_print(ERR_XIP_INTERNAL,"str_cat_fmt_va %s",xv->name);
	  xv++;
	  continue ;
	}
      if (w = XtNameToWidget(shell,fqvarname))
	{
	  char		*value;

	  XtVaGetValues(w,
			XtNstring,	&value,
			NULL);
	  if (strcmp(value,xip_var_get_static(xv->name)))
	    {
	      if ((status = xip_var_set(xv->name,value)) < 0)
		{
		  err_print(-status,"xip_var_set");
		}
	    }
	}
      else
	err_print(ERR_XIP_INTERNAL,"no such widget %s",fqvarname);
    }
  VEC_ENDFOR;
  if ((status = xip_cf_save(xiprc)) < 0)
    err_print(-status,"xip_cf_save %s",xiprc);
  XtDestroyWidget(shell);
}

t_xip_cf_box_entry		buffers_entries[] =
{
  {"tmplBufsizEntry",		"tmpl_bufsiz"},
  {"htmlBufiszEntry",		"html_bufsiz"},
  {"htmlizeBufsizEntry",	"htmlize_bufsiz"},
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_entry		layers_entries[] = 
{
  {"layDataBytesPerLineEntry",	"lay_data_bytes_per_line"},
  {"layUdpPerformSumEntry",	"lay_udp_perform_sum"},
  {"layUdpGuessModeEntry",	"lay_udp_guess_mode"},
  {"layerVerbLevelEntry",	"layer_verb_level"},
  {"layerBaseEntry",		"layer_base"},
  {"layPadNBytesEntry",		"lay_pad_nbytes"},
  {"layPadSubLayEntry",		"lay_pad_sub_lay"},
  {"etheraddrDecWayEntry",	"etheraddr_dec_way"},
  {"nbgethostTimeoutEntry",	"nbgethost_timeout"},
  {"xipDoOptsEntry",		"xip_do_opts"},
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_entry		ip_stack_entries[] = 
{
  {"ghostIpEntry",		"ghost_ip"},
  {"ghostEtherEntry",		"ghost_ether"},
  {"ipIdEntry",			"ip_id"},
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_entry		plugins_entries[] = 
{
  {"pluginsEntry",		"plugins"},
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_group		xip_cf_box_groups[] =
{
  {"buffersGroup",		"buffers",	buffers_entries},
  {"layersGroup",		"layers",	layers_entries},
  {"ipStackGroup",		"ip_stack",	ip_stack_entries},
  {"pluginsGroup",		"plugins",	plugins_entries},
  NULL_XIP_CF_BOX_GROUP
};

t_vec			*xip_cf_box_groups_vec = NULL;

t_status		xip_cf_box_group_add(xcbg)
t_xip_cf_box_group	*xcbg;
{
  t_xip_cf_box_group	*own_xcbg;
  t_status		status;
  
  if ((own_xcbg = XIP_ALLOC_PROC(sizeof (t_xip_cf_box_group),
			       "xip",
			       "xip_pkt_bar_add:ptr",
			       &status)) == NULL)
    return (status);
  FBCOPY(xcbg,own_xcbg,sizeof (t_xip_cf_box_group));
  if ((status = vec_add(xip_cf_box_groups_vec,
			own_xcbg)) < 0)
    {
      XIP_FREE_PROC(own_xcbg,
		    "xip",
		    "*:ptr");
      return (status);
    }
  return (0);
}

t_status		xip_cf_box_groups_init(VOID_DECL)
{
  t_status		status;
  t_xip_cf_box_group	*xcbg;

  if ((xip_cf_box_groups_vec = XIP_VEC_NEW(VEC_ONE_BASE,
				 &status)) == NULL)
    return (status);
  xcbg = xip_cf_box_groups;
  while (xcbg->widget_name)
    {
      if ((status = xip_cf_box_group_add(xcbg)) < 0)
	return (status);
      xcbg++;
    }
  return (0);
}

VOID_FUNC		xip_cf_box_groups_destroy(VOID_DECL)
{
  vec_ptr_delete(xip_cf_box_groups_vec);
}

/* creates a configuration panel from variables */
VOID_FUNC		XipCreateCfBox(parent)
Widget			parent;
{
  Widget		shell;
  t_status		status;
  Widget		cancelCommand;
  Widget		okCommand;
  Widget		tabs;

  INTERFACE(parent,
            GETCHILDSHELL(shell)
            ("xipCfBox",transientShellWidgetClass,
             XtNmappedWhenManaged,	False,
	     XtNtransientFor,		parent,
	     XtNallowShellResize,	True,
	     XtNiconPixmap,		xipicon_pixmap,
	     CHILD("form",formWidgetClass,
		   GETCHILD(tabs)
		   ("tabs",tabsWidgetClass,
		    INSERTCODE(
			      VEC_FOR(xip_cf_box_groups_vec,
				      t_xip_cf_box_group *xcbg)
			      {
				t_xip_cf_box_entry	*xcbe;
				
				INTERFACE(GETPARENT,
					  CHILD(xcbg->widget_name,
						groupWidgetClass,
						XtNnumCols,	2,
						XtNtabLabel,	xcbg->label,
						INSERTCODE(
						     xcbe = xcbg->entries;
						     while (xcbe->widget_name)
						     {
						       INTERFACE(GETPARENT,
								 CHILD(xcbe->widget_name,
								       labelWidgetClass,
								       XtNlabel,
								       xcbe->var,
								       END),
								 CHILD(xcbe->var,textfieldWidgetClass,
								       XtNstring,
								       xip_var_get_static(xcbe->var),
								       END),
								 ENDINTERFACE);
						       xcbe++;
						     }
						     ),
						END),
					  ENDINTERFACE);
			      }
			      VEC_ENDFOR;
			      ),
		    END),
		   GETCHILD(okCommand)
		   ("ok",commandWidgetClass,
		    XtNfromVert,	tabs,
		    CALLBACK(XtNcallback,
			     (XtCallbackProc)XipCfOkCommand,
			     shell),
		    END),
		   GETCHILD(cancelCommand)
		   ("cancel",commandWidgetClass,
		    XtNfromVert,	tabs,
		    XtNfromHoriz,	okCommand,
		    CALLBACK(XtNcallback,
			     (XtCallbackProc)XipCfCancelCommand,
			     shell),
		    END),
		   END),
	     REALIZE,
	     END),
	    ENDINTERFACE);
  XmgCenterWidget(shell);
  XmgSetDestroyOnDelete(shell);
  XtPopup(shell,XtGrabExclusive);
}
